/* compute pi using Monte Carlo method */

#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include "mpi.h"
#include "mpe.h"

#define DEBUG             0
#define CHUNKSIZE      1000
#define WINDOW_SIZE     500
#define QUARTER_CIRCLE    0
#define RANDOM_SEED       0

/* message tags */
#define NEED_NUMBERS      1
#define RANDOM_NUMBERS    2

DrawCircle( graph, centerx, centery, radius )
MPE_XGraph *graph;
int centerx, centery, radius;
{
  int lasty, x, y, radius2;

  radius2 = radius*radius;
  lasty = 0;
  for (x=1-radius; x<=radius; x++) {
    y = sqrt( (double)(radius2-x*x) );
    MPE_DrawLine( graph, x-1+centerx,  lasty+centery, x+centerx,  y+centery,
		 MPE_BLACK );
    MPE_DrawLine( graph, x-1+centerx, -lasty+centery, x+centerx, -y+centery,
		 MPE_BLACK );
    lasty = y;
  }
  MPE_Update( graph );
}
  


DrawArc( graph, size )
MPE_XGraph *graph;
int size;
{
  int lasty, x, y, size2;

  size2 = size*size;
  lasty = 0;
  for (x=1; x<=size; x++) {
    y = sqrt( (double)(size2-x*x) );
    MPE_DrawLine( graph, x-1, size-lasty, x, size-y,
		 MPE_BLACK );
    lasty = y;
  }
  MPE_Update( graph );
}
  


main(argc,argv)
int argc;
char *argv[];
{
    register int in,out,i,iters,max;
    register int ix,iy;
    register double x,y,calculatedPi;
    double error, epsilon;
    int ranks[1], done, temp;
    int numprocs, myid, randServer, totalin, totalout, workerNum;
    MPI_Comm world, workers;
    MPI_Group world_group, worker_group;
    MPI_Status mesgStatus;
    int randNums[CHUNKSIZE], request;
    MPE_XGraph *graph;

    MPI_Init(&argc,&argv);
    world = MPI_COMM_WORLD;
    MPI_Comm_size(world,&numprocs);
    MPI_Comm_rank(world,&myid);
    randServer = numprocs-1;

    if (myid == 0) {
      if ((argc!=2) || !sscanf( argv[1], "%lf", &epsilon )) {
	printf( "Syntax:\n    %s <epsilon>\n", argv[0] );
      }
    }
    MPE_OpenGraphics( &graph, MPI_COMM_WORLD, 0, -1, -1, WINDOW_SIZE,
		     WINDOW_SIZE, 0 );
    /*
       DrawArc( graph, WINDOW_SIZE );
       */

    MPI_Comm_group( world, &world_group );	/* get group of everyone  */
    ranks[0] = randServer;	                /* exclude the last       */
    MPI_Group_excl( world_group, 1, ranks, &worker_group );
                                                /* get new group    */
    MPI_Comm_create( world, worker_group, &workers );  /* get new comm     */
    

    if (myid == randServer)	/* I am the rand server */
    {
#if RANDOM_SEED
        struct timeval time;
        gettimeofday( &time, 0 );
	/* initialize the random number generator */
	srandom( (int)(time.tv_usec*1000000+time.tv_sec) );
#endif
	do {
	    MPI_Recv(&request, 1, MPI_INT, MPI_ANY_SOURCE, NEED_NUMBERS,
		     world, &mesgStatus);
	    if (request)
	    {
#if DEBUG
	        fprintf( stderr, "[%d] request from %d for numbers\n", myid,
			mesgStatus.MPI_SOURCE);
#endif
		for (i = 0; i < CHUNKSIZE; i++)
	            {
		        randNums[i] = random();
		    }
		MPI_Send(randNums, CHUNKSIZE, MPI_INT, mesgStatus.MPI_SOURCE, 
			 RANDOM_NUMBERS, world);
	    }
	}
	while( request>0 );
    }
    else			/* I am a worker process */
    {
        request = 1;
	done = in = out = 0;
	max = (1 << 31) - 1;	/* find max integer for normalization */
#if DEBUG
	fprintf( stderr, "[%d] asking for numbers\n", myid );
#endif
        MPI_Send( &request, 1, MPI_INT, randServer, NEED_NUMBERS, world );
	  /* request a string of random numbers */
        MPI_Comm_rank( workers, &workerNum );
	while (!done)
	{
	    request = 1;
	    MPI_Recv( randNums, CHUNKSIZE, MPI_INT, randServer, RANDOM_NUMBERS,
		     world, &mesgStatus );

#if DEBUG
	fprintf( stderr, "[%d] computing\n", myid );
#endif
	    for (i=0; i<CHUNKSIZE; )
	    {
#if QUARTER_CIRCLE
	        x = ((float) randNums[i++])/max;
		y = ((float) randNums[i++])/max;
#else
	        x = (((float) randNums[i++])/max) * 2 - 1;
		y = (((float) randNums[i++])/max) * 2 - 1;
#endif
#if DEBUG>1
		fprintf( stderr, "[%d] (%lf %lf)\n", myid, x, y );
#endif
		if (x*x + y*y < 1.0)
		{
		    in++;
#if QUARTER_CIRCLE
		    MPE_DrawPoint( graph,
				  (int)(x*WINDOW_SIZE), 
				  (int)(WINDOW_SIZE - y*WINDOW_SIZE),
				  MPE_BLACK );
#else
		    MPE_DrawPoint( graph,
				  (int)(WINDOW_SIZE/2 + x*WINDOW_SIZE/2), 
				  (int)(WINDOW_SIZE/2 - y*WINDOW_SIZE/2),
				  MPE_BLACK );
#endif
		}
		else
		    out++;
	    }
	    MPE_Update( graph );
	    temp = in;
#if DEBUG
	fprintf( stderr, "[%d] reducing\n", myid );
#endif
	    MPI_Reduce(&temp, &totalin, 1, MPI_INT, MPI_SUM, 0,
		       workers);
	      /* count total of ins  */ 
	    temp = out;
	    MPI_Reduce(&temp, &totalout, 1, MPI_INT, MPI_SUM, 0,
		       workers);
	      /* count total of outs */

	    if (myid == 0)
	    {
	        calculatedPi = (4.0*totalin)/(totalin + totalout);
		error = fabs( calculatedPi-3.141592653589793238462643);
		done = (error < epsilon) || (totalin+totalout)>1000000000;
		printf( "\rpi = %23.20lf", calculatedPi );
		request = (done) ? 0 : 1;
		MPI_Send( &request, 1, MPI_INT, randServer, NEED_NUMBERS,
			  world );
	        MPI_Bcast( &done, 1, MPI_INT, 0, workers );
	    } else
	    {
	        MPI_Bcast( &done, 1, MPI_INT, 0, workers );
		if (!done)
		{
		    request = 1;
		    MPI_Send( &request, 1, MPI_INT, randServer, NEED_NUMBERS,
			     world );
		}
	    }
	}
    }
    if (myid == 0)
    {
        printf( "\npoints: %d\nin: %d, out: %d\n", totalin+totalout,
	       totalin, totalout );
	getchar();
    }
    MPE_CloseGraphics( graph );
    MPI_Finalize();
}
