/*
 *  $Id: mpi++.h,v 1.3 1994/03/31 14:34:52 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/* user include file for MPI++ programs */

#ifndef _MPIPP_INCLUDE
#define _MPIPP_INCLUDE

extern "C" {
#include <mpi.h>
#include <mpi_proto.h>
int worker(int , char **);
}

/*
   Assumptions:

   MPI_Group is a pointer
   MPI_Comm is a pointer
*/

// Forward class declarations
extern class MPIX_Status;
extern class MPIX_Group;
extern class MPIX_Comm;
extern class MPIX_I_Comm_world;  // Only instance is MPI_COMM_WORLD 
extern class MPIX_I_Comm;
extern class MPIX_Intercomm;

typedef MPI_Op MPIX_Op;


//----------------------------------------------------------//
// MPIX_Status                                              //
//----------------------------------------------------------//
class MPIX_Status {
friend class MPIX_I_Comm;

public:
  // Constructor
  MPIX_Status (void) { }

  // Destructor
  ~MPIX_Status (void) { }

  // Accessors
  int Get_source(int& source_in)
	{ return (MPI_Get_source(status,source_in)); }
  int Get_tag(int& tag_in)
	{ return (MPI_Get_tag(status,tag_in)); }
  int Get_count(int& count_in)
	{ return (MPI_Get_count(status,count_in)); }

private:
  MPI_Status status;

  // Standard operations not provided by this class

  // New status objects cannot created from existing status objects
  // MPIX_Status(const MPIX_Status&) { }

  // Status objects cannot be assigned from other status objects
  // MPIX_Status& operator=(const MPIX_Status&) { }
};


//----------------------------------------------------------//
// MPIX_Group                                               //
//----------------------------------------------------------//
class MPIX_Group {
friend class MPIX_Comm;
friend class MPIX_I_Comm;

public:
  // Constructor
  MPIX_Group (void) { group = (MPI_Group)0; }

  // Destructor
  ~MPIX_Group (void) { Free(); }
  
  // Group accessors
  int Size(int& size_out) { return(MPI_Group_size ( group, &size_out )); }
  int Rank(int& rank_out) { return(MPI_Group_rank ( group, &rank_out)); }
  int Translate_ranks(int n, int r1[], MPIX_Group& group2, int r2[]) { 
	return (MPI_Group_translate_ranks(group, n, r1, group2.group, r2)); 
  }
  
  // Group constructors 
  int Union(MPIX_Group group2, MPIX_Group& group_out) {
	return (MPI_Group_union(group,group2.group,&(group_out.group)));
  }
  int Intersection(MPIX_Group group2, MPIX_Group& group_out) {
	return (MPI_Group_intersection(group,group2.group,&(group_out.group)));
  }
  int Difference(MPIX_Group group2, MPIX_Group& group_out) {
	return (MPI_Group_difference(group,group2.group,&(group_out.group)));
  }
  int Incl(int n, int ranks[], MPIX_Group& group_out) {
	return (MPI_Group_incl(group,n,ranks,&(group_out.group)));
  }
  int Excl(int n, int ranks[], MPIX_Group& group_out) {
	return (MPI_Group_excl(group,n,ranks,&(group_out.group)));
  }
  int Range_incl(int n, int ranges[], MPIX_Group& group_out) {
	return (MPI_Group_range_incl(group,n,ranges,&(group_out.group)));
  }
  int Range_excl(int n, int ranges[], MPIX_Group& group_out) {
	return (MPI_Group_range_excl(group,n,ranges,&(group_out.group)));
  }
  
  // Group destructors
  int Free(void) { return (MPI_Group_free(&group)); }

private:
  MPI_Group group;

  // Operations not provided by this class

  // Initialization of a group from an existing group is not allowed
  // MPIX_Group(const MPIX_Group&) { }

  // Cannot assign one group to another
  // MPIX_Group& operator=(const MPIX_Group&) { }
};


//----------------------------------------------------------//
// MPIX_I_Comm                                              //
//----------------------------------------------------------//
class MPIX_I_Comm {
friend class MPIX_Comm;
friend class MPIX_I_Comm_world;

public:
  // Constructor
  MPIX_I_Comm (void) { comm = (MPI_Comm)0; }

  // Destructor
  ~MPIX_I_Comm (void) { Free(); }

  // Accessors
  int Size(int& size_out) { 
	return(MPI_Comm_size ( comm, &size_out )); 
  }
  int Rank(int& rank_out) { 
	return(MPI_Comm_rank ( comm, &rank_out )); 
  }
  int Group(MPIX_Group& group_out) { 
	return (MPI_Comm_group(comm, &(group_out.group)));
  }
  
  // Initializers
  int Dup(MPIX_I_Comm& comm_out) {
	return (MPI_Comm_dup(comm, &(comm_out.comm)));
  }

  // Destructors
  int Free(void) { return (MPI_Comm_free(&comm)); }


  // Sends
  int Send(void *buf,int cnt, MPI_Datatype type, int dest, int tag) { 
	return (MPI_Send(buf,cnt,type,dest,tag,comm)); 
  }
  int Isend(void *buf, int cnt, MPI_Datatype type, int dest, int tag,
					MPI_Comm_request& request) { 
	return (MPI_Isend(buf,cnt,type,dest,tag,comm,&request)); 
  }


  // Receives
  int Recv(void *buf, int cnt, MPI_Datatype type, int source, 
				   int tag, MPIX_Status& status) { 
	return (MPI_Recv(buf,cnt,type,source,tag,comm, &(status.status))); 
  }
  int Irecv(void *buf, int cnt, MPI_Datatype type, int source, 
					int tag, MPI_Comm_request& request) { 
	return (MPI_Irecv(buf,cnt,type,source,tag,comm, &request)); 
  }
  

  // Completion operations
  int Wait(MPI_Comm_request& request, MPIX_Status& status) { 
	return (MPI_Wait(&request, &(status.status))); 
  }
  // Test needs to be fixed to be draft compliant
  int Test(MPI_Comm_request& request,int& flag,MPIX_Status& status) { 
	return (MPI_Test(request, &flag, &(status.status))); 
  }
  int Testany(int count, MPI_Comm_request reqs[], int& index, 
					  int& flag, MPIX_Status& status) {
	return (MPI_Testany(count, reqs, &index, &flag, &(status.status)));
  }
  int Waitall(int count, MPI_Comm_request reqs[], 
					  MPIX_Status status[]) {
	MPI_Status   stat[count];
	int error  = MPI_Waitall(count,reqs,stat);
	for(int i=0; i<count; i++) 
	  status[i].status    = stat[i];
	return (error);
  }
  // Need all of the persistent communication request functions
  // Also need the functions from 3.11

private:
  MPI_Comm comm;

  // Standard operations not provided by this class

  // Creating a new communicator from an existing communicator
  // MPIX_Comm(const MPIX_Comm&) { }
  
  // You cannot assign a value to MPIX_Comm's
  // MPIX_Comm& operator=(const MPIX_Comm&) { }
};


//----------------------------------------------------------//
// MPIX_Comm                                                //
//----------------------------------------------------------//
class MPIX_Comm: public MPIX_I_Comm {
public:
  // Constructor
  MPIX_Comm (void) { }

  // Initializers
  int Make(MPIX_Group& group, MPIX_Comm& comm_out) {
	return (MPI_Comm_make(comm, group.group, &(comm_out.comm)));
  }
  int Split(int color, int key, MPIX_Comm& comm_out) {
	return (MPI_Comm_split(comm, color, key, &(comm_out.comm)));
  }


  // Inter-communicator initializer
  int Intercomm_make(int ll, MPIX_Comm& pc, int rl, int tag, 
					 MPIX_Comm& comm_out) {
	return (MPI_Intercomm_make(comm,ll,pc.comm,rl,tag,&(comm_out.comm)));
  }


  // Collective operations
  int Barrier(void) { return (MPI_Barrier(comm)); }
  int Bcast(void *buf, int cnt, MPI_Datatype type, int root) { 
	return (MPI_Bcast(buf, cnt, type, root, comm)); 
  }
  int Gather(void *sbuf,int scnt, MPI_Datatype stype, void *rbuf,
					 int rcnt, MPI_Datatype rtype, int root) {
	return (MPI_Gather(sbuf,scnt,stype,rbuf,rcnt,rtype,root,comm)); 
  }
  int Gatherv(void *sbuf,int scnt, MPI_Datatype stype, void *rbuf,
					  int *rcnts, int *displs, MPI_Datatype rtype, int root) {
	return (MPI_Gatherv(sbuf,scnt,stype,rbuf,rcnts,displs,rtype,root,comm));
  }
  int Scatter(void *sbuf, int scnt, MPI_Datatype stype, void *rbuf,
					  int rcnt, MPI_Datatype rtype, int root) { 
	return (MPI_Scatter(sbuf,scnt,stype,rbuf,rcnt,rtype,root,comm)); 
  }
  int Scatterv(void *sbuf, int *scnts, int *displs, 
					   MPI_Datatype stype, void *rbuf, int rcnt, 
					   MPI_Datatype rtype, int root) {
	return (MPI_Scatterv(sbuf,scnts,displs,stype,rbuf,rcnt,rtype,root,
						 comm)); 
  }
  int Allgather(void *sbuf,int scnt, MPI_Datatype stype, void *rbuf,
						int rcnt, MPI_Datatype rtype) { 
	return (MPI_Allgather(sbuf,scnt,stype,rbuf,rcnt,rtype,comm)); 
  }
  int Allgatherv(void *sbuf,int scnt, MPI_Datatype stype, void *rbuf,
						 int *rcnts, int *displs, MPI_Datatype rtype) {
	return (MPI_Allgatherv(sbuf,scnt,stype,rbuf,rcnts,displs,rtype,comm)); 
  }
  int Alltoall(void *sbuf, int scnt, MPI_Datatype stype, void *rbuf,
					   int rcnt, MPI_Datatype rtype) {
	return (MPI_Alltoall(sbuf,scnt,stype,rbuf,rcnt,rtype,comm)); 
  }
  int Alltoallv(void *sbuf, int *scnts, int *sdispls, 
						MPI_Datatype stype, void *rbuf, int *rcnts, 
						int *rdispls, MPI_Datatype rtype) {
	return (MPI_Alltoallv(sbuf,scnts,sdispls,stype,rbuf,rcnts,
						  rdispls,rtype,comm)); 
  }
  int Reduce(void *in,void *out,int cnt,
					 MPI_Datatype type, MPIX_Op op, int root) {
	return (MPI_Reduce ( in, out, cnt, type, op, root, comm )); 
  }
  int Allreduce(void *in, void *out, int cnt, 
						MPI_Datatype type, MPI_Op op) {
	return (MPI_Allreduce (in, out, cnt, type, op, comm)); 
  }
  int Reduce_scatter(void *sbuf, void *rbuf, int *rcnts, 
							 MPI_Datatype type, MPI_Op op)	{ 
	return (MPI_Reduce_scatter(sbuf,rbuf,rcnts,type,op,comm)); 
  }
  int Scan(void *sbuf, void *rbuf, int cnt, 
				   MPI_Datatype type, MPI_Op op) {
	return (MPI_Scan(sbuf,rbuf,cnt,type,op,comm)); 
  }

  // Topology functions
  int Make_dims(int nnodes, int ndims, int dims[]) {
    return (MPI_Make_dims(nnodes,ndims,dims));
  }
  int Make_cart(int ndims, int dims[], int periods[], int reorder,
                MPIX_Comm& comm_out) {
    return (MPI_Make_cart(comm,ndims,dims,periods,reorder,&(comm_out.comm)));
  }
  int Cart_coords(int rank, int ndims, int coords[]) {
    return (MPI_Cart_coords(comm,rank,ndims,coords));
  }
  int Cart_rank(int coords[],int &rank) {
    return (MPI_Cart_rank(comm,coords,&rank));
  }
};


// MPI_COMM_WORLD - specific class
class MPIX_I_Comm_world : public MPIX_Comm {
public:
  int Init(int& argc, char **argv) {
	int error = MPI_Init (&argc, argv); 
	comm = MPI_COMM_WORLD;
	return (error);
  }
  int Finalize (void) { return (MPI_Finalize()); }

  
};


extern MPIX_I_Comm_world MPIX_COMM_WORLD;
extern const int MPIX_SUCCESS;

// Various defines to allow use of "MPI" instead of "MPIX" naming
#define MPI_COMM_WORLD MPIX_COMM_WORLD
#define MPI_Comm       MPIX_Comm
#define MPI_Group      MPIX_Group

#endif
