/**\ --MPE_Log--
*  * mpe_log.c - the externally callable functions in MPE_Log
*  *
*  * MPE_Log currently represents some code written by Dr. William
*  * Gropp, stolen from Chameleon's 'blog' logging package and
*  * modified by Ed Karrels, as well as some fresh code written
*  * by Ed Karrels.
*  *
*  * All work funded by Argonne National Laboratory
\**/

#include <stdio.h>

#define DEBUG 0
#define DEBUG_FORMATRECORD 0
#define DEBUG_RELOAD 0
#define DEBUG_LINKS 0

#include "mpi.h"
/* 
   Include a definition of MALLOC and FREE to allow the use of Chameleon
   memory debug code 
 */
#ifdef DEVICE_CHAMELEON
#include "mpisys.h"
#else
#define MALLOC(a) malloc(a)
#define FREE(a)   free(a)
#endif


#include "mpe_log.h"
#include "mpe_log_genproc.h"

#include "mpe_log_genproc.c"
#include "mpe_log_adjusttime.c"
#include "mpe_log_merge.c"

/*@
    MPE_Initlog - Initialize for logging
@*/
int MPE_Initlog()
{
  char fn[40];

  if (!MPE_Log_hasBeenInit || MPE_Log_hasBeenClosed) {
    if (!MPE_Log_Flush()) {	/* get the first block for data */
      fprintf(stderr, "MPE_Log: **** trace buffer creation failure ****\n");
      return MPE_Log_NO_MEMORY;
    }
    MPI_Comm_rank( MPI_COMM_WORLD, &MPE_Log_procid ); /* get process ID */
    MPE_Log_hasBeenInit = 1;	/* set MPE_Log as being initialized */
    MPE_Log_hasBeenClosed = 0;
    MPE_Log_isLockedOut = 0;
      /* set MPE_Log as being not closed, open for new data */
#if DEBUG
    sprintf( fn, "debug%d.out", MPE_Log_procid );
    debug_file = fopen( fn, "w" );
#endif
  }
  return MPE_Log_OK;
}





/*@
    MPE_Startlog - Begin logging of events
@*/
int MPE_Startlog()
{
  if (!MPE_Log_hasBeenInit) MPE_Initlog();
  MPE_Log_isLockedOut = 0;
  MPI_Barrier( MPI_COMM_WORLD );
  MPE_Log_init_clock();
  MPE_Log_event(MPE_Log_EVENT_SYNC,0,0);	/* log a 'sync' event */
  return MPE_Log_OK;
}



/*@
    MPE_Stoplog - Stop logging events
@*/
int MPE_Stoplog()
{
  MPE_Log_isLockedOut = 1;
  return MPE_Log_OK;
}




/*@
    MPE_Describe_state - Create log record describing a state

    Notes:
    Adds string containing a state def to the logfile.  The format of the
    def is (-13) 0 sevent eevent "color" "name".
@*/
int MPE_Describe_state( start, end, name, color )
int start, end;
char *name, *color;
{
  if (!MPE_Log_hasBeenInit) MPE_Initlog();
  if (MPE_Log_isLockedOut) return MPE_Log_LOCKED_OUT;
  MPE_Log_def_state (start, end, color, name);
  return MPE_Log_OK;
}



/*@
    MPE_Describe_event - Create log record describing an event type
@*/
int MPE_Describe_event( event, name )
int event;
char *name;
{
  if (!MPE_Log_hasBeenInit) MPE_Initlog();
  if (MPE_Log_isLockedOut) return MPE_Log_LOCKED_OUT;
  MPE_Log_def (event, name);
  return MPE_Log_OK;
}


/*@
    MPE_Log_event - Log an event
@*/
int MPE_Log_event(event,data,string)
int event, data;
char *string;
{
  MPE_Log_HEADER *b;
  MPE_Log_VFIELD *v;
  char c;
  int str_len,	    /* length of the string */
      str_store;    /* # of ints needed to store the string */
  double tempTime;  /* need to put here initally to avoid doubleword */
		          /* alignment problems */

  if (!MPE_Log_hasBeenInit) return MPE_Log_NOT_INITIALIZED;
  if (MPE_Log_isLockedOut) return MPE_Log_LOCKED_OUT;

  if (!string) {
      string	= "";
      str_len	= 0;
      str_store	= 0;
      }
  else {
      str_len	= strlen(string);
      str_store	= str_len / sizeof(int) + 1;
      }

#if DEBUG_LINKS
PrintBlockLinks( debug_file );
#endif

  tempTime=MPI_Wtime();
#if DEBUG>1
fprintf( stderr, "[%d] event %d at %10.5lf\n", MPE_Log_procid, event, tempTime );
#endif
  /*              header (w/event)     data                 string */
  if (MPE_Log_i + MPE_Log_HEADERSIZE + MPE_Log_VFIELDSIZE(1) + MPE_Log_VFIELDSIZE(str_store)
      > MPE_Log_size) {
    /* fprintf(stderr,"allocate more memory\n"); */
    if (!MPE_Log_Flush()) return MPE_Log_NO_MEMORY;
      /* cannot allocate more memory */
#if DEBUG
PrintBlockLinks( debug_file );
#endif
  }
  
  /*    MPE_Log_ADDHEADER (b, event); */
  b           = (MPE_Log_HEADER *)((int *)(MPE_Log_thisBlock+1) + MPE_Log_i);
  b->len      = (sizeof(MPE_Log_HEADER)/sizeof(int));
  b->event    = event;
  MOVEDBL( &b->time, &tempTime );
  MPE_Log_i   += b->len;
  
  MPE_Log_ADDINTS (b, v, 1, &data);	/* add one int to the definition */
  
  if (MPE_Log_i + MPE_Log_VFIELDSIZE(str_store) > MPE_Log_size) {
    
    /* fprintf(stderr,"long string cut, %d %d %d ints used for headers\n", 
       MPE_Log_HEADERSIZE, MPE_Log_VFIELDSIZE(1), MPE_Log_VFIELDSIZE(0)); */
    
    v           = (MPE_Log_VFIELD *)((int *)(MPE_Log_thisBlock+1) + MPE_Log_i);
    str_store   = MPE_Log_size - MPE_Log_i - MPE_Log_VFIELDSIZE(0);
    str_len     = str_store*sizeof( int )-1;
    v->len      = MPE_Log_VFIELDSIZE(str_store);  /* size of 'v' in ints */
    b->len      += v->len;
    v->dtype    = MPE_Log_CHAR;
    memcpy( v->other, string, str_len);
    ((char *)(v->other))[str_len]=0;   /* terminate string */
    MPE_Log_i   += v->len;
  } else {
    v           = (MPE_Log_VFIELD *)((int *)(MPE_Log_thisBlock+1) + MPE_Log_i);
    v->len      = MPE_Log_VFIELDSIZE(str_store);
    b->len      += v->len;
    v->dtype    = MPE_Log_CHAR;
    memcpy( v->other, string, str_len+1 );
    MPE_Log_i   += v->len;
  }
#if DEBUG
  fprintf( debug_file, "add event, record length %d, buffer size %d\n", b->len,
	   MPE_Log_i);
#endif
  return MPE_Log_OK;
}


/*@
    MPE_Finishlog - Send log to master, who writes it out

    Notes:
    This routine dumps a logfile in alog format
@*/
int MPE_Finishlog( filename )
char *filename;
{
  int returnStatus;
  if (!MPE_Log_hasBeenClosed) {
    MPI_Barrier(MPI_COMM_WORLD );
    MPE_Log_event(MPE_Log_EVENT_SYNC,0,0);	/* log a 'sync' event */
    MPE_Log_isLockedOut = 1;
    MPE_Log_thisBlock->size = MPE_Log_i;
#if DEBUG
    fprintf( debug_file, "adjusting times\n" );
    PrintBlockChain( debug_file, MPE_Log_firstBlock );
#endif
    MPE_Log_adjusttimes();
#if DEBUG
    fprintf( debug_file, "parallel merge\n" );
#endif
    if (returnStatus=MPE_Log_ParallelMerge( filename )) 
      return returnStatus;
    MPE_Log_hasBeenClosed = 1;
  }
  return MPE_Log_OK;
}

#undef DEBUG
#undef DEBUG_FORMATRECORD
#undef DEBUG_RELOAD
#undef DEBUG_LINKS
