/*
 *  $Id: comm_compare.c,v 1.3 1994/04/07 17:25:26 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>

/*@

MPI_Comm_compare - Compare two communicators

Input Parameters:
. comm1 - comm1 (handle) 
. comm2 - comm2 (handle) 
Output Parameter:
. result - integer which is MPI_IDENT if the contexts and groups are the
same, MPI_CONGRUENT if different contexts but identical groups, MPI_SIMILAR
if different contexts but similar groups, and MPI_UNEQUAL otherwise

@*/
int MPI_Comm_compare ( comm1, comm2, result )
MPI_Comm  comm1;
MPI_Comm  comm2;
int       *result;
{
  int       errno = MPI_SUCCESS;
  int       size1, size2;
  MPI_Group group1, group2;

  /* Check for bad arguments */
  if ( ( (comm1 == MPI_COMM_NULL) && (errno = MPI_ERR_COMM) ) ||
 	   ( (comm2 == MPI_COMM_NULL) && (errno = MPI_ERR_COMM) ) ||
	   ( (result == (int *)0)     && (errno = MPI_ERR_ARG) ) )
    return MPIR_ERROR( MPI_COMM_NULL, errno, 
					  "Error in MPI_COMM_COMPARE" );

  /* Are they the same kind of communicator */
  if (comm1->comm_type != comm2->comm_type) {
	(*result) = MPI_UNEQUAL;
	return (errno);
  }

  /* See if they are identical */
  if (comm1 == comm2) {
	(*result) = MPI_IDENT;
	return (errno);
  }
	
  /* Comparison for intra-communicators */
  if (comm1->comm_type == MPIR_INTRA) {

	/* Get the groups and see what their relationship is */
	MPI_Comm_group (comm1, &group1);
	MPI_Comm_group (comm2, &group2);
	MPI_Group_compare ( group1, group2, result );

	/* They can't be identical since they're not the same
	   handle, they are congruent instead */
	if ((*result) == MPI_IDENT)
	  (*result) = MPI_CONGRUENT;

	/* Free the groups */
	MPI_Group_free (&group1);
	MPI_Group_free (&group2);
  }
  
  /* Comparison for inter-communicators */
  else {
	int       lresult, rresult;
	MPI_Group rgroup1, rgroup2;
	
	/* Get the groups and see what their relationship is */
	MPI_Comm_group (comm1, &group1);
	MPI_Comm_group (comm2, &group2);
	MPI_Group_compare ( group1, group2, &lresult );

	MPI_Comm_remote_group (comm1, &rgroup1);
	MPI_Comm_remote_group (comm2, &rgroup2);
	MPI_Group_compare ( rgroup1, rgroup2, &rresult );

	/* Choose the result that is "least" strong. This works 
	   due to the ordering of result types in mpi.h */
	(*result) = (rresult > lresult) ? rresult : lresult;

	/* They can't be identical since they're not the same
	   handle, they are congruent instead */
	if ((*result) == MPI_IDENT)
	  (*result) = MPI_CONGRUENT;

	/* Free the groups */
	MPI_Group_free (&group1);
	MPI_Group_free (&group2);
	MPI_Group_free (&rgroup1);
	MPI_Group_free (&rgroup2);
  }

  return (errno);
}
