/*
 *  $Id: comm_dup.c,v 1.16 1994/04/13 21:43:39 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>
#include "mpisys.h"
#include "ic.h"

/*@

MPI_Comm_dup - Duplicates an existing communicator with all its cached
               information

Input Parameter:
. comm - communicator (handle) 
Output Parameter:
. newcomm - copy of comm (handle) 

@*/
int MPI_Comm_dup ( comm, comm_out )
MPI_Comm comm, *comm_out;
{
  MPI_Comm new_comm;

  /* Check for non-null communicator */
  if ( comm == MPI_COMM_NULL ) {
    (*comm_out) = MPI_COMM_NULL;
	return MPIR_ERROR( comm, MPI_ERR_COMM, "Error in MPI_COMM_DUP" );
  }

  /* Duplicate intra-communicators */
  if ( comm->comm_type == MPIR_INTRA ) {
	new_comm                  = NEW(MPIR_COMMUNICATOR);
	if (!new_comm) {
      return MPIR_ERROR( comm, MPI_ERR_EXHAUSTED, 
						"Out of space in MPI_COMM_DUP" );
	}
	new_comm->comm_type       = MPIR_INTRA;
	(void) MPIR_Group_dup ( comm->group, &(new_comm->group) );
	new_comm->local_group     = new_comm->group;
	(void) MPIR_Context_alloc ( comm, 2, &(new_comm->send_context) );
	new_comm->recv_context    = new_comm->send_context;
	new_comm->comm_cache      = 0;
	new_comm->topology.type   = comm->topology.type;
	new_comm->ref_count       = 1;
	new_comm->error_handler   = new_comm->error_handler;

	/* Copy topology info */
	if (comm->topology.type == MPI_CART) {
	  int i, ndims;
	  new_comm->topology.tag.cart.nnodes        = 
		comm->topology.tag.cart.nnodes; 
	  new_comm->topology.tag.cart.ndims = ndims = 
		comm->topology.tag.cart.ndims;
	  new_comm->topology.tag.cart.dims          = 
		(int *)MALLOC( sizeof(int) * 3 * ndims );
	  new_comm->topology.tag.cart.periods       = 
		new_comm->topology.tag.cart.dims + ndims;
	  new_comm->topology.tag.cart.position      = 
		new_comm->topology.tag.cart.periods + ndims;
	  for ( i=0; i<ndims; i++ ) {
		new_comm->topology.tag.cart.dims[i]     = 
		  comm->topology.tag.cart.dims[i];
		new_comm->topology.tag.cart.periods[i]  =
		  comm->topology.tag.cart.periods[i];
	  }
	  for ( i=0; i < ndims; i++ ) 
		new_comm->topology.tag.cart.position[i] =
		  comm->topology.tag.cart.position[i];
	}
	else if (comm->topology.type == MPI_GRAPH) {
	  int  i, nnodes;
	  int *index;
	  new_comm->topology.tag.graph.nnodes = nnodes = 
		comm->topology.tag.graph.nnodes;
	  new_comm->topology.tag.graph.nedges     = 
		comm->topology.tag.graph.nedges;
	  index = comm->topology.tag.graph.index;
	  new_comm->topology.tag.graph.index      = 
		(int *)MALLOC(sizeof(int) * (nnodes + index[nnodes-1]) );
	  new_comm->topology.tag.graph.edges      = 
		new_comm->topology.tag.graph.index + nnodes;
	  for ( i=0; i<nnodes; i++ )
		new_comm->topology.tag.graph.index[i] = 
		  comm->topology.tag.graph.index[i];
	  for ( i=0; i<index[nnodes-1]; i++ )
		new_comm->topology.tag.graph.edges[i] =
		  comm->topology.tag.graph.edges[i];
	}

	(void) MPIR_Attr_copy ( comm, new_comm ); 
	(void) MPIR_Comm_make_coll ( new_comm, MPIR_INTRA );
	(*comm_out)               = new_comm;
  }

  /* Duplicate inter-communicators */
  else {
	MPI_Comm     inter_comm = comm->comm_coll;
	MPI_Comm     intra_comm = comm->comm_coll->comm_coll;
	int          rank;
	MPIR_CONTEXT recv_context, send_context;

	/* Create the new inter-communicator */
	new_comm                  = NEW(MPIR_COMMUNICATOR);
	if (!new_comm) {
      return MPIR_ERROR( comm, MPI_ERR_EXHAUSTED, 
						"Out of space in MPI_COMM_DUP" );
	}
	new_comm->comm_type       = MPIR_INTER;
	(void) MPIR_Group_dup ( comm->group, &(new_comm->group) );
	(void) MPIR_Group_dup ( comm->local_group, &(new_comm->local_group) );

	/* Allocate send context, inter-coll context and intra-coll context */
	MPIR_Context_alloc ( intra_comm, 3, &recv_context );

	/* If I'm the local leader, then swap context info */
	MPI_Comm_rank ( intra_comm, &rank );
	if (rank == 0) {
	  MPI_Status status;
	  
	  MPI_Sendrecv (&recv_context, 1, MPIR_CONTEXT_TYPE, 0, MPIR_IC_DUP_TAG,
			&send_context, 1, MPIR_CONTEXT_TYPE, 0, MPIR_IC_DUP_TAG,
			inter_comm, &status);
	}
	
	/* Broadcast the send context */
	MPI_Bcast(&send_context, 1, MPIR_CONTEXT_TYPE, 0, intra_comm);

	/* We all now have all the information necessary,finish building the */
	/* inter-communicator */
	new_comm->send_context  = send_context;
	new_comm->recv_context  = recv_context;
	new_comm->comm_cache    = 0;
	new_comm->topology.type = MPI_UNDEFINED;
	new_comm->ref_count     = 1;	
	new_comm->error_handler = comm->error_handler;

	(void) MPIR_Attr_copy ( comm, new_comm ); 

	/* Build the collective inter-communicator */
	MPIR_Comm_make_coll( new_comm, MPIR_INTER );

	/* Build the collective intra-communicator */
	MPIR_Comm_make_coll ( new_comm->comm_coll, MPIR_INTRA );

	(*comm_out)               = new_comm;
  }

  return(MPI_SUCCESS);
}

