/*
 *  $Id: group_diff.c,v 1.8 1994/04/07 08:55:54 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>
#include "mpisys.h"

#define MPIR_UNMARKED 0
#define MPIR_MARKED   1

/*@

MPI_Group_difference - Make a group from the difference of two groups

Input Parameters:
. group1 - first group (handle) 
. group2 - second group (handle) 
Output Parameter:
. newgroup - difference group (handle) 

@*/
int MPI_Group_difference ( group1, group2, group_out )
MPI_Group group1, group2, *group_out;
{
  int        i, j, global_rank;
  MPI_Group  new_group;
  int        n;
  int        errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (group1 == MPI_GROUP_NULL) && (errno = MPI_ERR_GROUP) ) ||
 	   ( (group2 == MPI_GROUP_NULL) && (errno = MPI_ERR_GROUP) ) )
    return MPIR_ERROR( MPI_COMM_NULL, errno, 
					  "Error in MPI_GROUP_DIFFERENCE" );

  /* Check for EMPTY groups */
  if ( group1 == MPI_GROUP_EMPTY ) {
    MPIR_Group_dup ( MPI_GROUP_EMPTY, group_out );
    return (errno);
  }
  if ( group2 == MPI_GROUP_EMPTY ) {
    (void) MPIR_Group_dup ( group1, group_out );
    return (errno);
  }
  
  /* Create the new group */
  new_group = (*group_out) = NEW(MPIR_GROUP);
  new_group->ref_count     = 1;
  new_group->local_rank    = MPI_UNDEFINED;
  new_group->set_mark      = (int *)0;

  /* Set the number in the difference */
  n = group1->np;

  /* Allocate set marking space for group1 if necessary */
  if (group1->set_mark == NULL) {
    group1->set_mark = (int *) MALLOC( group1->np * sizeof(int) );
    if (!group1->set_mark) 
	  return MPIR_ERROR( MPI_COMM_NULL, MPI_ERR_EXHAUSTED, 
						"Out of space in MPI_GROUP_DIFFERENCE" );
  }

  /* Mark the difference */
  for ( i=0; i<group1->np; i++ ) {
    group1->set_mark[i] = MPIR_MARKED;
    for ( j=0; j<group2->np; j++ ) 
      if ( group1->lrank_to_grank[i] == group2->lrank_to_grank[j] ) {
        group1->set_mark[i] = MPIR_UNMARKED;
        n--;
        break;
      }
  }

  /* If there is a null intersection */
  if ( n <= 0 ) {
	FREE( new_group );
	MPIR_Group_dup ( MPI_GROUP_EMPTY, group_out );
	return (errno);
  }

  /* Alloc memory for lrank_to_grank array */
  new_group->np             = n;
  new_group->lrank_to_grank = (int *) MALLOC( n * sizeof(int) );
  if (!new_group->lrank_to_grank) {
      return MPIR_ERROR( MPI_COMM_NULL, MPI_ERR_EXHAUSTED, 
			 "Out of space in MPI_GROUP_DIFFERENCE" );
      }
    
  /* Fill in the space */
  for ( n=0, i=0; i<group1->np; i++ ) 
    if ( (group1->set_mark[i]==MPIR_MARKED) && (n < new_group->np) ) 
      new_group->lrank_to_grank[n++] = group1->lrank_to_grank[i];

  /* Find the local rank */
  MPID_myrank ( &global_rank );
  for( i=0; i<new_group->np; i++ )
    if ( global_rank == new_group->lrank_to_grank[i] ) {
      new_group->local_rank = i;
      break;
    }

  /* Determine the previous and next powers of 2 for this group */
  MPIR_Powers_of_2 ( new_group->np, &(new_group->N2_next), 
		     &(new_group->N2_prev) );

  return (errno);
}
