/*
 *  $Id: make_graph.c,v 1.8 1994/04/01 23:29:27 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>

/*@

MPI_Graph_create - Make a new communicator to which topology information
                 has been attached

Input Parameters:
. comm_old - input communicator without topology (handle) 
. nnodes - number of nodes in graph (integer) 
. index - array of integers describing node degrees (see below) 
. edges - array of integers describing graph edges (see below) 
. reorder - ranking may be reordered (true ) or not (false ) (logical) 

Output Parameter:
. comm_graph - communicator with graph topology added (handle) 

Algorithm:
We ignore the periods and reorder info currently.

@*/
int MPI_Graph_create ( comm_old, nnodes, index, edges, reorder, comm_graph )
MPI_Comm  comm_old;
int       nnodes;
int      *index;
int      *edges;
int       reorder;
MPI_Comm *comm_graph;
{
  int range[1][3];
  MPI_Group group_old, group;
  int i, rank, num_ranks = 1;
  int errno = MPI_SUCCESS;

  /* Check validity of arguments */
  if (((comm_old   == (MPI_Comm)0)   && (errno = MPI_ERR_COMM_NULL)) ||
      ((comm_graph == (MPI_Comm *)0) && (errno = MPI_ERR_NULL))      ||
      ((index      == (int *)0)      && (errno = MPI_ERR_NULL))      ||
      ((edges      == (int *)0)      && (errno = MPI_ERR_NULL))      ||
      ((nnodes     <  1)             && (errno = MPI_ERR_UNKNOWN))    )
    return MPIR_ERROR( comm_old, errno, "Error in MPI_GRAPH_CREATE" );
  
  /* Determine number of ranks in topology and make new comm */
  num_ranks = nnodes;
  if ( num_ranks < 1 ) {
    (*comm_graph)  = (MPI_Comm)0;
    return MPIR_ERROR( comm_old, MPI_ERR_TOPOLOGY, 
		       "Error in MPI_GRAPH_CREATE" );
  }
  range[0][0] = 0; range[0][1] = num_ranks - 1; range[0][2] = 1;
  if (errno = MPI_Comm_group ( comm_old, &group_old )) return errno;
  if (errno = MPI_Group_range_incl ( group_old, 1, range, &group )) 
      return errno;
  if (errno = MPI_Comm_make  ( comm_old, group, comm_graph )) return errno;

  /* Store topology information in new communicator */
  if ( (*comm_graph) != (MPI_Comm)0 ) {
    (*comm_graph)->topology.type                 = MPI_GRAPH;
    (*comm_graph)->topology.tag.graph.nnodes     = nnodes;
    (*comm_graph)->topology.tag.graph.nedges     = index[nnodes-1];
    (*comm_graph)->topology.tag.graph.index      = 
      (int *)MPIR_MALLOC( sizeof(int) * (nnodes + index[nnodes-1]) );
    (*comm_graph)->topology.tag.graph.edges      = 
      (*comm_graph)->topology.tag.graph.index + nnodes;
    for ( i=0; i<nnodes; i++ )
      (*comm_graph)->topology.tag.graph.index[i] = index[i];
    for ( i=0; i<index[nnodes-1]; i++ )
      (*comm_graph)->topology.tag.graph.edges[i] = edges[i];
  }
  return (errno);
}
